//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include <vector>
#include <set>

#include "io/VtkWriter.h"
#include "parallel/PetscProblemStat.h"
#include "parallel/PetscSolver.h"
#include "parallel/MpiHelper.h"

namespace AMDiS {

  using namespace std;


  PetscProblemStat::PetscProblemStat(string nameStr,
				     ProblemIterationInterface *problemIteration)
    : ParallelProblemStatBase(nameStr, problemIteration),
      deleteSolver(true)
  {
    FUNCNAME("PetscProblemStat::PetscProblemStat()");

    string name("");
    Parameters::get("parallel->solver", name);
    
    if (name == "petsc-schur") {
      petscSolver = new PetscSolverSchur();
    } else if (name == "petsc-feti") {
      petscSolver = new PetscSolverFeti();
    } else if (name == "petsc-block") {
      petscSolver = new PetscSolverGlobalBlockMatrix();
    } else if (name == "petsc" || name == "") {
      petscSolver = new PetscSolverGlobalMatrix();
    } else {
      ERROR_EXIT("No parallel solver %s available!\n", name.c_str());
    }
  }


  PetscProblemStat::PetscProblemStat(string nameStr,
				     PetscSolver *solver)
    : ParallelProblemStatBase(nameStr, NULL),
      petscSolver(solver),
      deleteSolver(false)
  {}


  void PetscProblemStat::initialize(Flag initFlag, 
				    ProblemStatSeq* adoptProblem,
				    Flag adoptFlag)
  {
      ParallelProblemStatBase::initialize(initFlag, adoptProblem, adoptFlag);

      meshDistributor->setBoundaryDofRequirement(petscSolver->getBoundaryDofRequirement());
  }


  void PetscProblemStat::solve(AdaptInfo *adaptInfo,
			       bool createMatrixData,
			       bool storeMatrixData)
  {
    FUNCNAME("PetscProblemStat::solve()");

    TEST_EXIT(meshDistributor)("Should not happen!\n");

    double wtime = MPI::Wtime();

#if 0
    double vm, rss;
    processMemUsage(vm, rss);       
    MSG("STAGE 1\n");
    MSG("My memory usage is VM = %.1f MB    RSS = %.1f MB\n", vm, rss);    
    mpi::globalAdd(vm);
    mpi::globalAdd(rss);
    MSG("Overall memory usage is VM = %.1f MB    RSS = %.1f MB\n", vm, rss);
#endif

    if (createMatrixData) {
      petscSolver->setMeshDistributor(meshDistributor);
      petscSolver->fillPetscMatrix(systemMatrix);
    }

    petscSolver->fillPetscRhs(rhs);

#if 0
    processMemUsage(vm, rss);   
    MSG("STAGE 2\n");
    MSG("My memory usage is VM = %.1f MB    RSS = %.1f MB\n", vm, rss);    
    mpi::globalAdd(vm);
    mpi::globalAdd(rss);
    MSG("Overall memory usage is VM = %.1f MB    RSS = %.1f MB\n", vm, rss);
#endif

    petscSolver->solvePetscMatrix(*solution, adaptInfo);   

    if (!storeMatrixData)
      petscSolver->destroyMatrixData();

#if 0
    processMemUsage(vm, rss);   
    MSG("STAGE 3\n");
    MSG("My memory usage is VM = %.1f MB    RSS = %.1f MB\n", vm, rss);    
    mpi::globalAdd(vm);
    mpi::globalAdd(rss);
    MSG("Overall memory usage is VM = %.1f MB    RSS = %.1f MB\n", vm, rss);
#endif

    INFO(info, 8)("solution of discrete system needed %.5f seconds\n", 
		  MPI::Wtime() - wtime);
  }

}
